# -*- coding: utf-8 -*-
#
import pygmsh
import examples

from importlib import import_module
import math
import numpy
import pytest
import voropy


def _get_volume(points, cells):
    cells = \
        cells['tetra'] if 'tetra' in cells \
        else cells['triangle']

    # Only points/cells that actually used
    uvertices, uidx = numpy.unique(cells, return_inverse=True)
    cells = uidx.reshape(cells.shape)
    points = points[uvertices]

    if cells.shape[1] == 4:
        mesh = voropy.mesh_tetra.MeshTetra(points, cells)
    else:
        assert cells.shape[1] == 3
        mesh = voropy.mesh_tri.MeshTri(points, cells)
    return math.fsum(mesh.cell_volumes)


@pytest.mark.parametrize('name', examples.__all__)
def test_check_output(name):
    test = import_module('examples.' + name)
    geom, vol = test.generate()
    points, cells, point_data, cell_data, _ = pygmsh.generate_mesh(geom)

    vol2 = _get_volume(points, cells)
    assert abs(vol - vol2) < 1.0e-3 * vol
    return



# -*- coding: utf-8 -*-
#


class Point(object):
    _POINT_ID = 0

    def __init__(self, x, lcar):
        self.x = x
        self.lcar = lcar

        self.id = 'p%d' % Point._POINT_ID
        Point._POINT_ID += 1

        fmt = ', '.join(len(x) * ['%r'])
        self.code = '\n'.join([
            '%s = newp;' % self.id,
            ('Point(%s) = {' + fmt + ', %r};')
            % ((self.id,) + tuple(x) + (lcar,))
            ])
        return



# -*- coding: utf-8 -*-
#
from .line_base import LineBase
from .point import Point


class CircleArc(LineBase):
    def __init__(self, start, center, end):
        super(CircleArc, self).__init__()

        assert isinstance(start, Point)
        assert isinstance(center, Point)
        assert isinstance(end, Point)

        self.start = start
        self.center = center
        self.end = end

        self.code = '\n'.join([
            '%s = newl;' % self.id,
            'Circle(%s) = {%s, %s, %s};'
            % (self.id, start.id, center.id, end.id)
            ])
        return



# -*- coding: utf-8 -*-
#
from .line_loop import LineLoop


class RuledSurface(object):
    _ID = 0

    def __init__(self, line_loop):
        assert isinstance(line_loop, LineLoop)

        self.line_loop = line_loop

        self.id = 'rs%d' % RuledSurface._ID
        RuledSurface._ID += 1

        self.code = '\n'.join([
            '%s = news;' % self.id,
            'Ruled Surface(%s) = {%s};' % (self.id, self.line_loop.id)
            ])
        return



# -*- coding: utf-8 -*-
#
from __future__ import print_function
import numpy
import voropy


def rotation_matrix(u, theta):
    '''Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    '''
    # Cross-product matrix.
    cpm = numpy.array([
        [0.0,   -u[2],  u[1]],
        [u[2],    0.0, -u[0]],
        [-u[1],  u[0],  0.0]
        ])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c \
        + s * cpm \
        + (1.0 - c) * numpy.outer(u, u)
    return R


def _is_flat(X, tol=1.0e-15):
    '''Checks if all points X sit in a plane.
    '''
    # find three points that don't sit on a line
    found = False
    for x2 in X:
        orth = numpy.cross(X[1] - X[0], x2 - X[0])
        orth_dot_orth = numpy.dot(orth, orth)
        if orth_dot_orth > tol:
            found = True
            break
    if not found:
        # All points even sit on a line
        return True
    norm_orth = numpy.sqrt(orth_dot_orth)
    norm_x_min_x0 = numpy.sqrt(numpy.einsum('ij, ij->i', X - X[0], X - X[0]))
    return (
        abs(numpy.dot(X - X[0], orth)) < tol * (1.0 + norm_orth*norm_x_min_x0)
        ).all()


def generate_mesh(
        geo_object,
        optimize=True,
        num_quad_lloyd_steps=10,
        num_lloyd_steps=1000,
        verbose=True,
        dim=3
        ):
    import meshio
    import os
    import subprocess
    import tempfile

    handle, filename = tempfile.mkstemp(suffix='.geo')
    os.write(handle, geo_object.get_code().encode())
    os.close(handle)

    handle, outname = tempfile.mkstemp(suffix='.msh')

    macos_gmsh_location = '/Applications/Gmsh.app/Contents/MacOS/gmsh'
    if os.path.isfile(macos_gmsh_location):
        gmsh_executable = macos_gmsh_location
    else:
        gmsh_executable = 'gmsh'

    cmd = [gmsh_executable, '-%d' % dim, filename, '-o', outname]
    if optimize:
        cmd += ['-optimize']
    if num_quad_lloyd_steps > 0:
        cmd += ['-optimize_lloyd', str(num_quad_lloyd_steps)]

    # http://stackoverflow.com/a/803421/353337
    p = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    if verbose:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            print(line.decode('utf-8'), end='')

    p.communicate()[0]
    if p.returncode != 0:
        raise RuntimeError(
            'Gmsh exited with error (return code %d).' %
            p.returncode
            )

    X, cells, pt_data, cell_data, field_data = meshio.read(outname)

    # Lloyd smoothing
    if not _is_flat(X) or 'triangle' not in cells:
        print(
            'Not performing Lloyd smoothing '
            '(only works for flat triangular meshes).'
            )
        return X, cells, pt_data, cell_data, field_data
    print('Lloyd smoothing...')
    # find submeshes
    a = cell_data['triangle']['geometrical']
    # http://stackoverflow.com/q/42740483/353337
    submesh_bools = {v: v == a for v in numpy.unique(a)}

    X, cells['triangle'] = voropy.smoothing.lloyd_submesh(
            X, cells['triangle'], submesh_bools,
            tol=0.0, max_steps=num_lloyd_steps,
            verbose=False
            )

    return X, cells, pt_data, cell_data, field_data
